# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"filter_files": "00_core.ipynb",
         "ls": "00_core.ipynb",
         "hdf_attr_check": "00_core.ipynb",
         "dict2json": "00_core.ipynb",
         "Path.ls": "06_cli.ipynb",
         "monthlen": "00_core.ipynb",
         "InOutPath": "00_core.ipynb",
         "ProjectPath": "00_core.ipynb",
         "Region": "01_geo.ipynb",
         "open_shp": "01_geo.ipynb",
         "open_tif": "01_geo.ipynb",
         "bounds_from_shapefile": "01_geo.ipynb",
         "size_from_bounds": "01_geo.ipynb",
         "size_resolution_assert": "01_geo.ipynb",
         "rasterize": "01_geo.ipynb",
         "downsample": "01_geo.ipynb",
         "is_intersection": "01_geo.ipynb",
         "polygon_from_bounds": "01_geo.ipynb",
         "crop": "04_predict.ipynb",
         "bounds_from_coords": "01_geo.ipynb",
         "VIIRS375_download": "02_data.ipynb",
         "VIIRS750_download": "02_data.ipynb",
         "BandsFilter": "02_data.ipynb",
         "BandsRename": "02_data.ipynb",
         "MergeTiles": "02_data.ipynb",
         "BandsAssertShape": "02_data.ipynb",
         "ActiveFiresLog": "02_data.ipynb",
         "InterpolateAngles": "02_data.ipynb",
         "MirCalc": "02_data.ipynb",
         "BaseDataset": "02_data.ipynb",
         "mask2nan": "02_data.ipynb",
         "nan2zero": "02_data.ipynb",
         "group_files": "02_data.ipynb",
         "ViirsDataset": "02_data.ipynb",
         "Viirs375Dataset": "02_data.ipynb",
         "Viirs750Dataset": "02_data.ipynb",
         "MCD64Dataset": "02_data.ipynb",
         "FireCCI51Dataset": "02_data.ipynb",
         "AusCoverDataset": "02_data.ipynb",
         "MTBSDataset": "02_data.ipynb",
         "ICNFDataset": "02_data.ipynb",
         "Region2Tiles": "02_data.ipynb",
         "DROP": "03_models.ipynb",
         "BTNorm2d": "03_models.ipynb",
         "LSTM": "03_models.ipynb",
         "SpaceConv": "03_models.ipynb",
         "UpSpaceConv": "03_models.ipynb",
         "TimeConv": "03_models.ipynb",
         "UpTimeConv": "03_models.ipynb",
         "SpaceTimeConv": "03_models.ipynb",
         "UpSpaceTimeConv": "03_models.ipynb",
         "BA_Net": "03_models.ipynb",
         "open_nc": "04_predict.ipynb",
         "image2tiles": "04_predict.ipynb",
         "tiles2image": "04_predict.ipynb",
         "get_preds": "04_predict.ipynb",
         "predict_one": "04_predict.ipynb",
         "save_netcdf": "04_predict.ipynb",
         "inspect_netcdf": "04_predict.ipynb",
         "load_netcdf": "04_predict.ipynb",
         "predict_time": "04_predict.ipynb",
         "predict_month": "04_predict.ipynb",
         "predict_nrt": "04_predict.ipynb",
         "split_mask": "04_predict.ipynb",
         "RunManager": "04c_historical.ipynb",
         "banet_viirs375_download": "06_cli.ipynb",
         "banet_viirs750_download": "06_cli.ipynb",
         "banet_create_dataset": "06_cli.ipynb",
         "banet_dataset2tiles": "06_cli.ipynb",
         "banet_predict_monthly": "06_cli.ipynb",
         "banet_predict_times": "06_cli.ipynb",
         "banet_nrt_run": "06_cli.ipynb",
         "array2png": "07_web.ipynb"}

modules = ["core.py",
           "geo.py",
           "data.py",
           "models.py",
           "predict.py",
           "nrt.py",
           "historical.py",
           "cli.py",
           "web.py"]

doc_url = "https://mnpinto.github.io/banet/"

git_url = "https://github.com/mnpinto/banet/tree/master/"

def custom_doc_links(name): return None
